﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using System.Threading;
using Microsoft.ServiceModel.WebSockets;

namespace CallWCFWS
{
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.PerSession, ConcurrencyMode = ConcurrencyMode.Multiple)]
    public class WSChat:WebSocketsService
    {    
        private static ChatSessions sessions = new ChatSessions();
        private static int globalId;

        private int id;

        /// <summary>
        /// Initializes a new instance of the <see cref="WSChat"/> class.
        /// </summary>
        public WSChat()
        {
            this.id = Interlocked.Increment(ref WSChat.globalId);
            if (!WSChat.sessions.TryAdd(this))
            {
                throw new InvalidOperationException("Can't add session.");
            }
        }

        public override void OnMessage(string value)
        {
            WSChat.sessions.RelayMessage(this, value);
        }

        /// <summary>
        /// Returns the hash code for this instance
        /// </summary>
        /// <returns>A 32-bit signed integer hash code.</returns>
        //public override int GetHashCode()
        //{
        //    return this.id;
        //}

        protected override void OnClose(object sender, EventArgs e)
        {
            WSChat.sessions.Remove(this);
        }
    }

    internal class ChatSessions : IDisposable
    {
        private List<WSChat> innerCache = new List<WSChat>();
        private ReaderWriterLockSlim thisLock = new ReaderWriterLockSlim();

        /// <summary>
        /// Relaying the contents from one session to all other sessions.
        /// </summary>
        /// <param name="current">The current session relaying the content.</param>
        /// <param name="value">Content to relay.</param>
        public void RelayMessage(WSChat current, string value)
        {
            List<WSChat> defunct = null;
            //this.thisLock.EnterReadLock();
                foreach (var entry in this.innerCache)
                {
                        entry.SendMessage(value);
                }
                //this.thisLock.ExitReadLock();
        }

        /// <summary>
        /// セッション管理
        /// </summary>
        /// <param name="entry">Session to add.</param>
        /// <returns>true if session was added; false otherwise.</returns>
        public bool TryAdd(WSChat entry)
        {
            this.thisLock.EnterUpgradeableReadLock();
            try
            {
                if (this.innerCache.Contains(entry))
                {
                    return false;
                }

                this.thisLock.EnterWriteLock();
                try
                {
                    this.innerCache.Add(entry);
                    return true;
                }
                finally
                {
                    this.thisLock.ExitWriteLock();
                }
            }
            finally
            {
                this.thisLock.ExitUpgradeableReadLock();
            }
        }

        /// <summary>
        /// Attempting to remove a session from the collection.
        /// </summary>
        /// <param name="entry">Session to remove.</param>
        public void Remove(WSChat entry)
        {
            this.innerCache.Remove(entry);
        }

        /// <summary>
        /// Disposes the instance.
        /// </summary>
        public void Dispose()
        {
            this.thisLock.Dispose();
        }
    }
}